package modelage;

import java.util.Random;

import org.jgap.FitnessFunction;
import org.jgap.IChromosome;

/**
 * defines fitness function that is used in genetic algorithm
 * the larger the function, the better the solution is
 */

public class MODELAGEFitnessFunction extends FitnessFunction {	
	private static final long serialVersionUID = 1L;
	
	public static double[] initialWtValues;
	
	private double[] initialArKValues;
	private int noOfInputValues = 0;

	private boolean defaultComparation;
	private double trialKArValueDiag;
	private double trialKArValueDetr;
	private double trialKdetrPerKdiag;
	private double inverse;
	private double fitness;
	private double[] ikValue; //wt values corrected for Kdetr/Kdiag and further corrected for systematic error
	private double[] chromosomeArPerKValue;
	
	/**
	 * constructor reads, copies input variables and initializes
	 */
	public MODELAGEFitnessFunction(double[] initial_Ar_K_Values, double[] initial_Wt_Values, int no_Of_Input_Values) {
		noOfInputValues = no_Of_Input_Values;
		initialArKValues = new double[no_Of_Input_Values]; //experimental values
		initialWtValues = new double[no_Of_Input_Values]; //experimental values
	
		ikValue = new double[noOfInputValues];
		chromosomeArPerKValue = new double[noOfInputValues];
		
	//reading and copying of variables	
		for (int i=0; i<no_Of_Input_Values; i++){
			initialArKValues[i] = initial_Ar_K_Values[i];
			initialWtValues[i] = initial_Wt_Values[i];

		//correction of 40Ar/40K values is necessary if age of detritial 1M component is known	
			if (MODELAGE.systematicAgeKnown == true){
    			initialArKValues[i] = ArKCorrectToDetritialComponent (initialArKValues[i], MODELAGE.systematicError, MODELAGE.systematicAge);
    		}
		} 
	}

	/**
	 * Determine how good the solution is	
	 */
    public double evaluate(IChromosome a_subject) {
    // Take care of the fitness evaluator.
    	defaultComparation = a_subject.getConfiguration().getFitnessEvaluator().isFitter(2, 1);
    	trialKArValueDiag = getValueOfChromosome (a_subject, 0);
    	trialKArValueDetr = getValueOfChromosome (a_subject, 1);
    	trialKdetrPerKdiag = getValueOfChromosome (a_subject, 2);

    	fitness = 0; // value that describes property of the solution
    
    // Determine fitness of the solution
    	if (defaultComparation) {
    		inverse = 0;
    		for (int i=0; i<noOfInputValues; i++){
    			ikValue[i] = IkCalculate(initialWtValues[i], trialKdetrPerKdiag);//correction for Kdetr/Kdiag
    			ikValue[i] = addNoise(ikValue[i], 0, MODELAGE.systematicError);//correction for systematic error
    			chromosomeArPerKValue[i] = ArPerKTotal (ikValue[i], trialKArValueDiag, trialKArValueDetr);
    			inverse += (initialArKValues[i] - chromosomeArPerKValue[i])*(initialArKValues[i] - chromosomeArPerKValue[i]);	
    		}

    	//normalization
    	//	inverse = 1.0/noOfInputValues * Math.sqrt(inverse);

    		fitness = 1000 * 1/(1+50*inverse);
    	}
    	return fitness;
    }
 
	/**
	 * returns value of particular chromosome	
	 */
	public static double getValueOfChromosome (IChromosome a_potentialSolution,	int a_position) {
		Double ValueOfChromosome = (Double) a_potentialSolution.getGene(a_position).getAllele();
		return ValueOfChromosome.doubleValue();
	}

	/**
	 * modifies %wt values - here in this class it is used only to add systematic_Noise_Range
	 */
	private static double addNoise(double initial_Wt_Value, double random_Noise_Range, double systematic_Noise_Range){
		Random rand = new Random();
		double step = 0;
		boolean sign;
		double finalValue = 0;
			        
		step = rand.nextDouble();
		sign = rand.nextBoolean();
	   
		if (MODELAGE.systematicAgeKnown == true){
			if (sign == true){ //plus
				finalValue =  initial_Wt_Value + step * random_Noise_Range;
			}else {//minus
				finalValue =  initial_Wt_Value - step * random_Noise_Range;
			}
		} else {
			if (sign == true) //plus
				finalValue =  initial_Wt_Value + step * random_Noise_Range + (1 - initial_Wt_Value) * systematic_Noise_Range/100;
			else //minus
				finalValue =  initial_Wt_Value - step * random_Noise_Range + (1 - initial_Wt_Value) * systematic_Noise_Range/100;
		}
		return finalValue;
	}
	
	/**
	 * calculates Ik if wt. detr and Kdetrital/Kdiagenetic ratio is known
	 */
	private static double IkCalculate(double detritic_percent, double KdetPerKdiag){
		double Ik = 0;

		Ik = (detritic_percent * KdetPerKdiag)/(detritic_percent * KdetPerKdiag + (100-detritic_percent));
		  
		return Ik; // w zakresie 0-1
	} 
	
    /**
     * corrects 40Ar/40K values if 1Md detrital age is known 
     */
    private static double ArKCorrectToDetritialComponent (double ArKTotal, double systematicError, double systematicAge){
    	return (100 * ArKTotal - systematicError * ArKCalculate(systematicAge))/(100 - systematicError);
    }

	/**
	 * calculates 40Ar/40K for given Ik if end member values are known	
	 */
    private static double ArPerKTotal (double Ik, double ArPerK_diag, double ArPerK_detr){
		return Ik * ArPerK_detr + (1 - Ik) * ArPerK_diag;
	}	
	
    /**
     * calculates 40Ar/40K from age (in Ma)
     */
    private static double ArKCalculate(double age) {
    	return (Math.exp(age/10000*5.543) - 1)/1.4263 * 100;
    }
}///:~





